from django.conf import settings
from django.core.management.base import NoArgsCommand
from django.utils.importlib import import_module

from django_rpy2.models import AvailableLibrary

import sys

class Command(NoArgsCommand):
    help = "Can be run as a cronjob or directly to install any scheduled libraries."

    def handle_noargs(self, **options):
        for lib in AvailableLibrary.objects.filter(scheduled=True):
            sys.stderr.write("Installing R Library: %s" % str(lib))
            sys.stderr.flush()
            sys.stderr.write(" (%s)\n" % str(lib.install()))
        AvailableLibrary.objects.refresh()




try:
    from django.conf.urls import patterns, url, include
except ImportError:
    from django.conf.urls.defaults import patterns, url, include

from .views import *


urlpatterns = patterns('',
#  url(r'^$',                   Scripts(), name="charts"),
#  url(r'^(?P<slug>[\w-]+)/$',  Script(), name="chart"),
)




import datetime
import tempfile
import shutil

from django.utils import timezone
from django.test import TestCase

from .core import *
from .models import *
from .settings import *

TEST_DOWNLOADS = False

class RunnerTests(TestCase):
    def setUp(self):
        self.script = UploadedScript.objects.create(
            name="Test Script",
            rsc="foo")

    def test_script_variables(self):
        self.assertScript('1', foo=1)
        self.assertScript('Fox', foo="Fox")
        self.assertScript('NULL', foo=None)
        self.assertScript('1', foo=self.script)

    def assertScript(self, value, **kwargs):
        result = self.script.run(**kwargs)
        self.assertEqual(result.result, value)


class CoreTests(TestCase):
    def _pre_setup(self):
        if TEST_DOWNLOADS:
            self.libs = tempfile.mktemp('-test-libs')
        else:
            self.libs = '/tmp/rpy2-tests'
        self.r = get_R(libs=self.libs)
        return super(CoreTests, self)._pre_setup()

    def _post_teardown(self):
        if os.path.isdir(self.libs):
            shutil.rmtree(self.libs)
        return super(CoreTests, self)._post_teardown()

    def test_get_r(self):
        """Get R"""
        pass

    def test_directory(self):
        """Lib directory is added"""
        self.assertEqual(self.r('.libPaths()')[0], self.libs)

    def test_mirror(self):
        """CRAN mirror is set"""
        ret = str(self.r('getOption("repos")'))
        self.assertEqual(ret.strip().split()[-1].strip('"'), CRAN_MIRROR)

    def test_run(self):
        """Running a script"""
        ret = ScriptRunner("'1'").run()
        self.assertEqual(ret[0], '1')
        runner = ScriptRunner()
        self.assertEqual(runner.run("'2'")[0], '2')

    def test_require(self):
        """Import an R library"""
        runner = ScriptRunner('my - unrun - script')
        with self.assertRaises(ImportError):
            runner.libs.append('doesnotexist')
        self.assertNotIn('doesnotexist', str(runner))
        runner.libs.append('class')
        self.assertIn('require(class)\n', str(runner))

    def test_install(self):
        """Installing library"""
        libs = Library()
        if TEST_DOWNLOADS:
            self.assertFalse(libs.install('four-candles'))
            self.assertTrue(libs.install('ref'))

    def test_output(self):
        """Adding output filename"""
        runner = ScriptRunner('my - unrun - script')
        runner.use_output('fn_two', '/tmp/test')
        self.assertIn('/tmp/test', str(runner))

    def test_output_extension(self):
        """Getting output file with extension"""
        runner = ScriptRunner()
        runner.run("""
           x <- matrix(1:10, ncol = 5)
           write(x, paste(filename,'.txt',sep=""), sep="\t")
        """)
        self.assertTrue(os.path.isfile(runner.filename))
        with open(runner.filename, 'r') as fhl:
            self.assertEqual(fhl.read(), '1\t2\t3\t4\t5\n6\t7\t8\t9\t10\n')

        os.unlink(runner.filename)

    def test_csv_file(self):
        """Using CSV files"""
        filename = tempfile.mktemp('.csv')
        with open(filename, 'w') as fhl:
            fhl.write("""A,B,C\n1,2,3\n4,5,6\n""")
        runner = ScriptRunner()
        runner.use_csvfile('csv', filename)
        ret = runner.run("csv")
        self.assertEqual(ret[0][1], 4)
        os.unlink(filename)

    def test_varialbles(self):
        """Variables"""
        runner = ScriptRunner()
        ret = runner.run('foo', foo='bar\'"}]+=-')
        self.assertEqual(ret[0], 'bar\'"}]+=-')
        ret = runner.run('bar', bar=1)
        self.assertEqual(ret[0], 1)
        ret = runner.run('huh', huh=None)
        self.assertEqual(str(ret), 'NULL')

    def test_database(self):
        runner = ScriptRunner()
        if TEST_DOWNLOADS:
            with self.assertRaises(ImportError):
                runner.use_database('default')
        else:
            try:
                runner.use_database('default')
            except ImportError:
                runner.libs.install('RSQLite')
                runner.use_database('default')
        #self.assertIn('dbConnect', str(runner))
        #ret = runner.run("dbListTables(default)")
        # TODO: In future we want to try and connect to the testcase test database
        # We can't do that here because django has made the database 'in-memory'
        # and when we connect, R created a new 'in-memory' database instead of seeing
        # django's in-memory database.
        #self.assertIn('auth_user', tables)

    




from django.contrib.admin import *
from django.utils.safestring import mark_safe
from django_object_actions import DjangoObjectActions

from .models import *
from .forms import LibraryForm

def schedule_reinstall(modeladmin, request, queryset):
    queryset.update(installed=None, scheduled=True, attempted=None)
schedule_reinstall.short_description = "Re-schedule installation or upgrade"

class LibraryAdmin(ModelAdmin):
    list_display = ('__str__', 'is_installed', 'location', 'installed', 'scheduled', 'attempted')
    ordering = 'installed',
    actions = [ schedule_reinstall ]
    readonly_fields = ['installed', 'location', 'attempted', 'version']
    form = LibraryForm


class ResultsInline(TabularInline):
    exclude = ['fn_out']
    fields = ['render_url', 'run_ok', 'started', 'ended', 'output', 'result']
    readonly_fields = ['render_url', 'run_ok', 'output', 'result', 'started', 'ended']
    model = ScriptResult
    extra = 0
    max_num = 0

    def render_url(self, obj):
        if obj.fn_out:
            return mark_safe("""<a href="%s">%s</a>""" % (obj.fn_out.url, obj.fn_out.name))
        return "No File"

    def run_ok(self, obj):
        return not obj.error
    run_ok.boolean = True


class UploadAdmin(DjangoObjectActions, ModelAdmin):
    filter_horizontal = ('libs',)
    inlines = [ ResultsInline ]
    exclude = ['slug']
    readonly_fields = []
    objectactions = ('run', 'download')

    class Media:
        css = { 'all' : ('css/no-inline-label.css',) }

    def run(self, request, obj):
        obj.run()

    run.label = "Run Script"
    run.short_description = "Run this script right now (no save)"

    def download(self, request, obj):
        obj.run(save_instead=True)

    download.label = "Download Script"
    download.short_description = "Compile the script as it would be sent to R and output as the result instead of running."


site.register(AvailableLibrary, LibraryAdmin)
site.register(UploadedScript, UploadAdmin)
site.register(ScriptResult)




"""
Provide all the default access to R (via rpy2)
"""

import os
import sys
import tempfile

from glob import glob
from django.utils.text import slugify as _sl
slugify = lambda t: _sl(unicode(t))

from django.db.models import Model
from .settings import *

__all__ = ('get_R', 'Packages', 'Library', 'ScriptRunner')

try:
    from rpy2.robjects import r, rinterface, conversion
    from rpy2.rinterface import NULL
    
    def get_r_command():
        return r
except:
    # This code has been copied in without much testing if you
    # need pyRserve, please text, fix and submit a patch.
    import pyRserve
    conn = None
    rServeHost = 'localhost'
    rServePort = 6311

    def get_r_command():
        global conn
        if conn and type(conn) is pyRserve.rconn.RConnector and not conn.isClosed:
            return conn.r
        conn = pyRserve.connect(host=rServeHost, port=6311)
        conn.r.__call__ = conn.eval
        return conn.r

R = None

def write_out(self, name):
    def _inner(msg):
        setattr(self, name, getattr(self, name, '') + msg)
    return _inner

def no_input(msg):
    raise IOError("R is asking for something on the command line. Intervention required!")

def r_set(name, value):
    robj = conversion.py2ri(value)
    rinterface.globalenv.__setitem__(name, robj)

def get_R(libs=LIB_PATH, mirror=CRAN_MIRROR):
    global R
    if R is None:
        R = get_r_command()

        R.out, R.err = ('', '')
        R.output = write_out(R, 'out')
        rinterface.set_writeconsole(R.output)
        rinterface.set_readconsole(no_input)
        R.set_value = r_set

        if not os.path.isdir(libs):
            os.makedirs(libs)

        # RPY2 can popup GUI menus, disable this.
        R('options(menu.graphics=FALSE)')

        # Install library destination (and location for imports)
        R('.libPaths( c( "%s", .libPaths()) )' % (libs.rstrip('/')+"/"))

        # Setup install source repository using RPY_CRAN_MIRROR
        R("""local({r <- getOption("repos");
           r["CRAN"] <- "%s"; 
           options(repos=r)})""" % mirror)
    return R

class Packages(list):
    def __iter__(self):
        if not len(self):
            self.generate()
        return super(Packages, self).__iter__()

    def generate(self):
        """Return a list of available packages"""
        r = get_R()
        packages = r('available.packages()[,"Package"]')
        versions = r('available.packages()[,"Version"]')
        for x, name in enumerate(packages):
            self.append( (name, "%s (%s)" % (name, versions[x])) )

    def installed(self):
        """
        Look for installed packages,

        Returns: triplet of name, version and directory location.
        """
        packages  = r('installed.packages()[,1]')
        locations = r('installed.packages()[,2]')
        versions  = r('installed.packages()[,3]')
        for x, name in enumerate(packages):
            yield (name, versions[x], locations[x])

# Note: rpy2 provides package management code, but it provides
# very little benefit over doing it ourselves.
class Library(list):
    def append(self, lib):
        cmd = 'require(%s)' % lib
        if cmd not in self:
            if not list(get_R()(cmd))[0]:
                raise ImportError("Can't import required R module: '%s'" % lib)
            super(Library, self).append(cmd)

    def install(self, lib):
        """Attempts to install a library using R's"""
        r = get_R()
        r.out = ''
        r("install.packages(\"%s\")" % lib)
        # There is no way to know for sure if this worked, rpy2 has
        # a certain limit in what it will get back from R and this
        # Stops us known /for sure/ what's going on.
        return 'downloaded' in r.out.strip()

    def uninstall(self, lib):
        r = get_R()
        r.out = ''
        try:
            r("remove.packages(\"%s\")" % lib)
        except Exception:
            return None
        return True


class ScriptRunner(list):
    """
    Run an R script with start-up and wrap-up sections.
    """
    DB_ENGINE_TO_R_DB = {
       'mysql': ('RMySQL', """
library("RMySQL")
%(DB)s <- dbConnect(MySQL(), user="%(USER)s", password="%(PASSWORD)s", dbname="%(NAME)s", host="%(HOST)s", port="%(PORT)s")
"""),
       'sqlite3': ('RSQLite', """
library(DBI)
%(DB)s <- dbConnect(RSQLite::SQLite(), "%(NAME)s", flags=SQLITE_RO)
"""),
       'postgresql_psycopg2': ('RPostgreSQL', """
library(RPostgreSQL)
%(DB)s <- dbConnect(dbDriver("PostgreSQL"), user="%(USER)s", password="%(PASSWORD)s", dbname="%(NAME)s", host="%(HOST)s", port="%(PORT)s")
""")}

    def __init__(self, script=None, libs=None):
        self.r = get_R()
        self.libs = libs or Library()
        self.close = [ ]
        self.script = [ script or '' ]
        self.filename = None

    def use_output(self, name, filename):
        if name == 'filename':
            self.filename = filename
        if not os.path.isdir(os.path.dirname(filename)):
            raise IOError("Directory for file doesn't exist: %s" % filename)
        if os.path.isfile(filename):
            raise IOError("Refusing to overwrite file: %s" % filename)
        self.append( "%s = \"%s\"" % (slugify(name), filename))

    def use_csvfile(self, name, filename):
        if not os.path.isfile(filename):
            raise IOError("Can't load csv file: %s" % filename)
        self.append( "%s = read.csv(\"%s\")" % (slugify(name), filename))

    def use_database(self, name):
        try:
            db = DATABASES[name]
        except KeyError:
            raise KeyError("Database '%s' not found." % name)

        engine = db['ENGINE'].split('.')[-1]
        r_db = self.DB_ENGINE_TO_R_DB.get(engine, None)
        if not r_db:
            raise KeyError("Database type '%s' not supported." % engine)

        # Add the required package
        try:
            self.libs.append(r_db[0])
        except ImportError:
            return False

        args = db.copy()
        args['DB'] = slugify(name)

        self.append( self.DB_ENGINE_TO_R_DB[engine][-1] % args )
        self.close.append( "dbDisconnect(%s)" % args['DB'] )
        return True

    def __str__(self):
        return "\n\n".join( self + self.libs + self.script + self.close )

    def run(self, script=None, save_instead=False, **variables):
        if not self.filename:
            self.use_output('filename', tempfile.mktemp())

        if script:
            self.script = [ script ]

        for (name, value) in variables.items():
            if isinstance(value, Model):
                value = value.pk
            if isinstance(value, (str, unicode, int, float, bool)):
                self.r.set_value(str(name), value)
            else:
                self.r.set_value(str(name), NULL)

        self.r.out = ''
        if save_instead:
            with open(self.filename + '.r', 'w') as fhl:
                fhl.write(str(self))
            ret = '[Script Not Run]'
        else:
            ret = self.r(str(self))

        if not os.path.isfile(self.filename):
            self.filename = (glob(self.filename + '.*') + [self.filename])[0]

        return ret



from django.db.models import *
from django.utils.text import slugify
from django.utils.timezone import now
from django.core.files.base import File
from django.dispatch import receiver

from rpy2.robjects.vectors import Vector

import os
import sys

from .settings import DATABASES, LIB_PATH
from .core import *

__all__ = ('AvailableLibrary', 'UploadedScript', 'ScriptResult')

class LibraryManager(Manager):
    def refresh(self):
        for (name, version, location) in AvailableLibrary.PACKAGES.installed():
            goc = self.get_or_create(name=name, location=location,
                    defaults=dict(installed=now(), scheduled=False))
            goc[0].version = version
            goc[0].save()

class AvailableLibrary(Model):
    PACKAGES  = Packages()

    name      = CharField(max_length=32, choices=PACKAGES)
    version   = CharField(max_length=12, null=True, blank=True)
    location  = CharField(max_length=255, default="unknown")
    installed = DateTimeField(null=True, blank=True)
    attempted = DateTimeField(null=True, blank=True)
    scheduled = BooleanField("Schedule Install", default=True)

    objects = LibraryManager()

    class Meta:
        verbose_name_plural = 'Available Libraries'
        unique_together = ('name', 'location')

    def __str__(self):
        (a, b) = (bool(self.installed), bool(self.attempted))
        if self.location[0] == '/' and \
           self.location.rstrip('/') != LIB_PATH.rstrip('/'):
            msg = 'system lib'
        else:
            msg = [['not installed', 'failed install'],
                   ['manual install', 'installed']][a][b]

        if self.scheduled:
            msg = 'scheduled'

        version = self.version + ' ' if self.version else ''
        return "%s (%s%s)" % (self.name, version, msg)

    def is_installed(self):
        return bool(self.installed)
    is_installed.boolean = True

    def install(self):
        self.attempted = now()
        self.scheduled = False
        if Library().install(self.name):
            self.installed = now()
            self.location = LIB_PATH
        self.save()
        return bool(self.installed)

    def uninstall(self, keep=True):
        Library().uninstall(self.name)
        if keep:
            self.installed = None
            self.location = 'deleted'
            self.save()


@receiver(signals.pre_delete, sender=AvailableLibrary, dispatch_uid='dellib')
def delete_lib(sender, instance, using, **kwargs):
    instance.uninstall(keep=False)


DB_CHOICES = [ (db, db) for db in DATABASES.keys() ]

class UploadedScript(Model):
    name  = CharField(max_length=64)
    slug  = SlugField(max_length=64)
    libs  = ManyToManyField(AvailableLibrary, null=True, blank=True, limit_choices_to={'installed__isnull': False})
    db    = CharField('Database', max_length=32, choices=DB_CHOICES, null=True, blank=True)

    rsc   = TextField('R Script', help_text="""Use data sources:

      Uploaded CSV Data is available as 'csv'
      Website databases available as 'default' or name of database.
      Output filename is 'filename'.
    """)
    csv    = FileField('CSV Data', upload_to='rpy/csv/', null=True, blank=True)

    def __str__(self):
        return self.name

    def save(self, **kwargs):
        if not self.slug:
            self.slug = slugify(unicode(self.name))
        return super(UploadedScript, self).save()

    def get_libs(self):
        libs = Library()
        for lib in self.libs.filter(installed__isnull=False):
            try:
                libs.append(lib.name)
            except ImportError:
                lib.installed = None
                lib.save()
        return libs

    def run(self, **variables):
        result = self.results.create()
        result.started = now()
        result.save()

        runner = ScriptRunner(libs=self.get_libs())

        if self.db:
            runner.use_database(self.db)

        try:
            ret = runner.run(str(self.rsc).replace('\r',''), **variables)
        except Exception as error:
            result.error = True
            result.result = str(error)
        else:
            if isinstance(ret, Vector):
                if len(ret) == 1:
                    ret = str(ret[0])
                else:
                    ret = str(list(ret))
            else:
                ret = str(ret)
            result.result = ret

        if os.path.isfile(runner.filename):
            with open(runner.filename, 'r') as fhl:
                result.fn_out.save(os.path.basename(runner.filename), File(fhl))
                os.unlink(runner.filename)

        result.output = runner.r.out
        result.ended = now()
        result.save()

        return result


class ScriptResult(Model):
    script = ForeignKey(UploadedScript, related_name='results')

    started = DateTimeField(null=True, blank=True)
    ended   = DateTimeField(null=True, blank=True)

    fn_out = FileField('File Output', upload_to='rpy/out/', null=True, blank=True)
    output = TextField('Printed to Screen', null=True, blank=True)
    result = TextField('Result or Error', null=True, blank=True)
    error  = BooleanField(default=False)

    class Meta:
        ordering = ('-started',)

    def __str__(self):
        return "Script run for '%s' on '%s'" % (str(self.script), str(self.started))






import os

try:
    from django.conf import settings
except ImportError:
    settings = None

# Can be overloaded with RPY_DATABASES (instead of using django's db)
DATABASES = settings.DATABASES

# Can be overloaded with RPY_LIB_PATH
LIB_PATH = os.path.join(settings.MEDIA_ROOT, 'rpy', 'libs')

# Can be overloaded with RPY_CRAN_MIRROR
CRAN_MIRROR = "http://cran.us.r-project.org"

# Get all our overloaded settings and all of django's globals
for (name, default) in locals().items():
    locals()[name] = getattr(settings, 'RPY_'+name, default)





# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_rpy2', '0002_scriptresult_error'),
    ]

    operations = [
        migrations.AddField(
            model_name='availablelibrary',
            name='location',
            field=models.CharField(default=b'unknown', max_length=255),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='availablelibrary',
            name='name',
            field=models.CharField(max_length=32),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='availablelibrary',
            unique_together=set([('name', 'location')]),
        ),
    ]



#!/usr/bin/env python

import os
import sys

from setuptools import setup, find_packages

# remove MANIFEST. distutils doesn't properly update it when the contents of directories change.
if os.path.exists('MANIFEST'): os.remove('MANIFEST')

# Grab description for Pypi
with open('README') as fhl:
    description = fhl.read()

setup(
        name             = 'django-rpy2',
        version          = '0.6.6',
        description      = 'Allow django to create R scripts and generate charts and other data output from those scripts.',
        long_description = description,
        url              = 'https://bitbucket.net/charts',
        author           = 'Martin Owens',
        author_email     = 'doctormo@gmail.com',
        platforms        = 'linux',
        license          = 'AGPLv3',
        packages         = find_packages(),
        install_requires = [
          'django>=1.7.4',
          'django-object-actions>=0.5.1',
          'rpy2>=2.5.6',
        ],
    )




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AvailableLibrary',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(unique=True, max_length=32)),
                ('version', models.CharField(max_length=12, null=True, blank=True)),
                ('installed', models.DateTimeField(null=True, blank=True)),
                ('attempted', models.DateTimeField(null=True, blank=True)),
                ('scheduled', models.BooleanField(default=True, verbose_name=b'Schedule Install')),
            ],
            options={
                'verbose_name_plural': 'Available Libraries',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ScriptResult',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('started', models.DateTimeField(null=True, blank=True)),
                ('ended', models.DateTimeField(null=True, blank=True)),
                ('fn_out', models.FileField(upload_to=b'rpy/out/', null=True, verbose_name=b'File Output', blank=True)),
                ('output', models.TextField(null=True, verbose_name=b'Printed to Screen', blank=True)),
                ('result', models.TextField(null=True, verbose_name=b'Result or Error', blank=True)),
            ],
            options={
                'ordering': ('-started',),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='UploadedScript',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=64)),
                ('slug', models.SlugField(max_length=64)),
                ('db', models.CharField(blank=True, max_length=32, null=True, verbose_name=b'Database', choices=[(b'default', b'default')])),
                ('rsc', models.TextField(help_text=b"Use data sources:\n\n      Uploaded CSV Data is available as 'csv'\n      Website databases available as 'default' or name of database.\n      Output filename is 'filename'.\n    ", verbose_name=b'R Script')),
                ('csv', models.FileField(upload_to=b'rpy/csv/', null=True, verbose_name=b'CSV Data', blank=True)),
                ('libs', models.ManyToManyField(to='django_rpy2.AvailableLibrary', null=True, blank=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='scriptresult',
            name='script',
            field=models.ForeignKey(related_name='results', to='django_rpy2.UploadedScript'),
            preserve_default=True,
        ),
    ]



from django.contrib.admin import *
from django.forms import ModelForm, ChoiceField

from .models import *

class LateChoiceField(ChoiceField):
    def _get_choices(self):
        return self._choices

    def _set_choices(self, value):
        self._choices = self.widget.choices = value

    choices = property(_get_choices, _set_choices)

class LibraryForm(ModelForm):
    name = LateChoiceField(choices=AvailableLibrary.PACKAGES)

